"use strict";
var Fiber = require("fibers");

var iconv = require("iconv-lite");
var os = require("os");
var osenv = require("osenv");
var path = require("path");

var MobileHelper = require("./../mobile-helper");
var options = require("../../options");
var helpers = require("../../helpers");

var AndroidEmulatorServices = (function () {
    function AndroidEmulatorServices($logger, $emulatorSettingsService, $errors, $childProcess, $fs, $staticConfig) {
        this.$logger = $logger;
        this.$emulatorSettingsService = $emulatorSettingsService;
        this.$errors = $errors;
        this.$childProcess = $childProcess;
        this.$fs = $fs;
        this.$staticConfig = $staticConfig;
        iconv.extendNodeEncodings();
    }
    AndroidEmulatorServices.prototype.checkAvailability = function () {
        var _this = this;
        return (function () {
            var platform = MobileHelper.DevicePlatforms[1 /* Android */];
            if (!_this.$emulatorSettingsService.canStart(platform).wait()) {
                _this.$errors.fail("The current project does not target Android and cannot be run in the Android emulator.");
            }
        }).future()();
    };

    AndroidEmulatorServices.prototype.startEmulator = function (app, emulatorOptions) {
        var _this = this;
        return (function () {
            var image = options.avd || _this.getBestFit().wait();

            if (image) {
                _this.startEmulatorCore(app, emulatorOptions.appId, image).wait();
            } else {
                _this.$errors.fail("Could not find an emulator image to run your project.");
            }
        }).future()();
    };

    AndroidEmulatorServices.prototype.startEmulatorCore = function (app, appId, image) {
        var _this = this;
        return (function () {
            var initiallyRunningEmulators = _this.getRunningEmulators().wait();

            if (initiallyRunningEmulators.length === 0 || options.avd) {
                _this.$logger.info("Starting Android emulator with image %s", image);
                _this.$childProcess.spawn('emulator', ['-avd', image], { stdio: ["ignore", "ignore", "ignore"], detached: true }).unref();
            }

            var runningEmulators = _this.getRunningEmulators().wait();
            var endTimeEpoch = helpers.getCurrentEpochTime() + _this.getMilliSecondsTimeout();
            var hasTimeLeft = helpers.getCurrentEpochTime() < endTimeEpoch;

            while (hasTimeLeft) {
                if (runningEmulators.length > initiallyRunningEmulators.length || (runningEmulators.length > 0 && !options.avd)) {
                    break;
                }

                _this.$logger.trace("Restarting adb server");
                _this.sleep(10000);
                _this.$childProcess.spawnFromEvent(_this.$staticConfig.adbFilePath, ["kill-server"], "exit").wait();
                _this.sleep(1000);
                runningEmulators = _this.getRunningEmulators().wait();
                hasTimeLeft = helpers.getCurrentEpochTime() < endTimeEpoch;
            }

            if (!hasTimeLeft) {
                _this.$errors.fail(AndroidEmulatorServices.UNABLE_TO_START_EMULATOR_MESSAGE);
            }

            var emulatorId = _.first(runningEmulators);
            if (options.avd) {
                _.forEach(runningEmulators, function (emulator) {
                    if (!_.contains(initiallyRunningEmulators, emulator)) {
                        emulatorId = emulator;
                    }
                });
            }

            _this.waitForEmulatorBootToComplete(emulatorId, endTimeEpoch).wait();

            var childProcess = _this.$childProcess.spawn(_this.$staticConfig.adbFilePath, ["-s", emulatorId, "shell", "input", "keyevent", "82"]);
            _this.$fs.futureFromEvent(childProcess, "close").wait();

            _this.$logger.info("installing %s through adb", app);
            childProcess = _this.$childProcess.spawn(_this.$staticConfig.adbFilePath, ["-s", emulatorId, 'install', '-r', app]);
            _this.$fs.futureFromEvent(childProcess, "close").wait();

            _this.$logger.info("running %s through adb", app);
            childProcess = _this.$childProcess.spawn(_this.$staticConfig.adbFilePath, ["-s", emulatorId, 'shell', 'am', 'start', '-S', appId + "/" + _this.$staticConfig.START_PACKAGE_ACTIVITY_NAME], { stdio: ["ignore", "ignore", "ignore"], detached: true });
            _this.$fs.futureFromEvent(childProcess, "close").wait();
        }).future()();
    };

    AndroidEmulatorServices.prototype.sleep = function (ms) {
        var fiber = Fiber.current;
        setTimeout(function () {
            return fiber.run();
        }, ms);
        Fiber.yield();
    };

    AndroidEmulatorServices.prototype.getMilliSecondsTimeout = function () {
        var timeout = AndroidEmulatorServices.TIMEOUT_SECONDS;

        if (options && options.timeout) {
            var parsedValue = parseInt(options.timeout);
            if (!isNaN(parsedValue) && parsedValue > 0) {
                timeout = parsedValue;
            } else {
                this.$logger.info("timeout should be number bigger than 0. Default value: " + timeout + " seconds will be used.");
            }
        }

        return timeout * 1000;
    };

    AndroidEmulatorServices.prototype.getRunningEmulators = function () {
        var _this = this;
        return (function () {
            var emulatorDevices = [];
            var outputRaw = _this.$childProcess.execFile(_this.$staticConfig.adbFilePath, ['devices']).wait().split(os.EOL);
            _.each(outputRaw, function (device) {
                var rx = device.match(/^(emulator-\d+)\s+device$/);
                if (rx && rx[1]) {
                    emulatorDevices.push(rx[1]);
                }
            });
            return emulatorDevices;
        }).future()();
    };

    AndroidEmulatorServices.prototype.getBestFit = function () {
        var _this = this;
        return (function () {
            var minVersion = _this.$emulatorSettingsService.minVersion;

            var best = _.chain(_this.getAvds().wait()).map(function (avd) {
                return _this.getInfoFromAvd(avd).wait();
            }).max(function (avd) {
                return avd.targetNum;
            }).value();

            return (best.targetNum >= minVersion) ? best.name : null;
        }).future()();
    };

    AndroidEmulatorServices.prototype.getInfoFromAvd = function (avdName) {
        var _this = this;
        return (function () {
            var iniFile = path.join(_this.avdDir, avdName + ".ini");
            var avdInfo = _this.parseAvdFile(avdName, iniFile).wait();
            if (avdInfo.path && _this.$fs.exists(avdInfo.path).wait()) {
                iniFile = path.join(avdInfo.path, "config.ini");
                avdInfo = _this.parseAvdFile(avdName, iniFile, avdInfo).wait();
            }
            return avdInfo;
        }).future()();
    };

    AndroidEmulatorServices.prototype.parseAvdFile = function (avdName, avdFileName, avdInfo) {
        var _this = this;
        if (typeof avdInfo === "undefined") { avdInfo = null; }
        return (function () {
            var encoding = _this.getAvdEncoding(avdFileName).wait();
            var contents = _this.$fs.readText(avdFileName, encoding).wait().split("\n");

            avdInfo = _.reduce(contents, function (result, line) {
                var parsedLine = line.split("=");
                var key = parsedLine[0];
                switch (key) {
                    case "target":
                        result.target = parsedLine[1];
                        result.targetNum = _this.readTargetNum(result.target);
                        break;
                    case "path":
                        result.path = parsedLine[1];
                        break;
                    case "hw.device.name":
                        result.device = parsedLine[1];
                        break;
                    case "abi.type":
                        result.abi = parsedLine[1];
                        break;
                    case "skin.name":
                        result.skin = parsedLine[1];
                        break;
                    case "sdcard.size":
                        result.sdcard = parsedLine[1];
                        break;
                }
                return result;
            }, avdInfo || Object.create(null));
            avdInfo.name = avdName;
            return avdInfo;
        }).future()();
    };

    AndroidEmulatorServices.prototype.readTargetNum = function (target) {
        var platform = target.replace('android-', '');
        var platformNumber = +platform;
        if (isNaN(platformNumber)) {
            if (platform === "L") {
                platformNumber = 20;
            }
        }
        return platformNumber;
    };

    AndroidEmulatorServices.prototype.getAvdEncoding = function (avdName) {
        var _this = this;
        return (function () {
            var encoding = "utf8";
            var contents = _this.$fs.readText(avdName, "ascii").wait();
            if (contents.length > 0) {
                contents = contents.split("\n", 1)[0];
                if (contents.length > 0) {
                    var matches = contents.match(AndroidEmulatorServices.ENCODING_MASK);
                    if (matches) {
                        encoding = matches[1];
                    }
                }
            }
            return encoding;
        }).future()();
    };

    Object.defineProperty(AndroidEmulatorServices.prototype, "androidHomeDir", {
        get: function () {
            return path.join(osenv.home(), AndroidEmulatorServices.ANDROID_DIR_NAME);
        },
        enumerable: true,
        configurable: true
    });

    Object.defineProperty(AndroidEmulatorServices.prototype, "avdDir", {
        get: function () {
            return path.join(this.androidHomeDir, AndroidEmulatorServices.AVD_DIR_NAME);
        },
        enumerable: true,
        configurable: true
    });

    AndroidEmulatorServices.prototype.getAvds = function () {
        var _this = this;
        return (function () {
            var result = [];
            if (_this.$fs.exists(_this.avdDir).wait()) {
                var entries = _this.$fs.readDirectory(_this.avdDir).wait();
                result = _.select(entries, function (e) {
                    return e.match(AndroidEmulatorServices.INI_FILES_MASK) !== null;
                }).map(function (e) {
                    return e.match(AndroidEmulatorServices.INI_FILES_MASK)[1];
                });
            }
            return result;
        }).future()();
    };

    AndroidEmulatorServices.prototype.waitForEmulatorBootToComplete = function (emulatorId, endTime) {
        var _this = this;
        return (function () {
            helpers.printInfoMessageOnSameLine("Waiting for emulator device initialization...");

            while (helpers.getCurrentEpochTime() < endTime) {
                var isEmulatorBootCompleted = _this.isEmulatorBootCompleted(emulatorId).wait();

                if (isEmulatorBootCompleted) {
                    helpers.printInfoMessageOnSameLine(os.EOL);
                    return;
                }

                helpers.printInfoMessageOnSameLine(".");
                _this.sleep(10000);
            }

            helpers.printInfoMessageOnSameLine(os.EOL);
            _this.$errors.fail(AndroidEmulatorServices.UNABLE_TO_START_EMULATOR_MESSAGE);
        }).future()();
    };

    AndroidEmulatorServices.prototype.isEmulatorBootCompleted = function (emulatorId) {
        var _this = this;
        return (function () {
            var output = _this.$childProcess.execFile(_this.$staticConfig.adbFilePath, ["-s", emulatorId, "shell", "getprop", "dev.bootcomplete"]).wait();
            var matches = output.match("1");
            return matches && matches.length > 0;
        }).future()();
    };
    AndroidEmulatorServices.ANDROID_DIR_NAME = ".android";
    AndroidEmulatorServices.AVD_DIR_NAME = "avd";
    AndroidEmulatorServices.INI_FILES_MASK = /^(.*)\.ini$/i;
    AndroidEmulatorServices.ENCODING_MASK = /^avd\.ini\.encoding=(.*)$/;
    AndroidEmulatorServices.TIMEOUT_SECONDS = 120;
    AndroidEmulatorServices.UNABLE_TO_START_EMULATOR_MESSAGE = "Cannot run your app in the native emulator. Increase the timeout of the operation with the --timeout option. Alternatively, run the Android Virtual Device manager and increase the allocated RAM for the virtual device.";
    return AndroidEmulatorServices;
})();
$injector.register("androidEmulatorServices", AndroidEmulatorServices);
