var jaroWinklerDistance = require("../vendor/jaro-winkler_distance");
var helpers = require("./../helpers");
var util = require("util");

var options = require("../options");

var CommandArgumentsValidationHelper = (function () {
    function CommandArgumentsValidationHelper(isValid, remainingArguments) {
        this.isValid = isValid;
        this.remainingArguments = remainingArguments;
    }
    return CommandArgumentsValidationHelper;
})();

var CommandsService = (function () {
    function CommandsService($errors, $logger, $injector, $staticConfig, $hooksService) {
        this.$errors = $errors;
        this.$logger = $logger;
        this.$injector = $injector;
        this.$staticConfig = $staticConfig;
        this.$hooksService = $hooksService;
    }
    CommandsService.prototype.allCommands = function (includeDev) {
        var commands = this.$injector.getRegisteredCommandsNames(includeDev);
        return _.reject(commands, function (command) {
            return _.contains(command, '|');
        });
    };

    CommandsService.prototype.executeCommandUnchecked = function (commandName, commandArguments) {
        var _this = this;
        return (function () {
            var command = _this.$injector.resolveCommand(commandName);
            if (command) {
                if (!command.disableAnalytics) {
                    var analyticsService = _this.$injector.resolve("analyticsService");
                    analyticsService.checkConsent(commandName).wait();
                    analyticsService.trackFeature(commandName).wait();
                }
                if (command.enableHooks === undefined || command.enableHooks === true) {
                    var childrenCommandsNames = _.map(_this.$injector.getChildrenCommandsNames(commandName), function (command) {
                        return command.replace("*", "");
                    });
                    if (_.contains(childrenCommandsNames, commandArguments[0])) {
                        commandName = util.format("%s-%s", commandName, commandArguments[0]);
                    }

                    _this.$hooksService.initialize(commandName);
                    _this.$hooksService.executeBeforeHooks().wait();
                    command.execute(commandArguments).wait();
                    _this.$hooksService.executeAfterHooks().wait();
                } else {
                    command.execute(commandArguments).wait();
                }
                return true;
            }
            return false;
        }).future()();
    };

    CommandsService.prototype.executeCommand = function (commandName, commandArguments) {
        return this.executeCommandAction(commandName, commandArguments, this.executeCommandUnchecked);
    };

    CommandsService.prototype.executeCommandAction = function (commandName, commandArguments, action) {
        var _this = this;
        return this.$errors.beginCommand(function () {
            return action.apply(_this, [commandName, commandArguments]);
        }, function () {
            return _this.executeCommandUnchecked("help", [_this.beautifyCommandName(commandName)]);
        });
    };

    CommandsService.prototype.tryExecuteCommand = function (commandName, commandArguments) {
        var _this = this;
        return (function () {
            if (_this.executeCommandAction(commandName, commandArguments, _this.canExecuteCommand).wait()) {
                _this.executeCommandAction(commandName, commandArguments, _this.executeCommandUnchecked).wait();
            }
        }).future()();
    };

    CommandsService.prototype.canExecuteCommand = function (commandName, commandArguments) {
        var _this = this;
        return (function () {
            var command = _this.$injector.resolveCommand(commandName);
            var beautifiedName = helpers.stringReplaceAll(commandName, "|", " ");
            if (command) {
                if (command.canExecute) {
                    return command.canExecute(commandArguments).wait();
                }

                if (_this.$injector.isValidHierarchicalCommand(commandName, commandArguments)) {
                    return true;
                }

                if (_this.validateCommandArguments(command, commandArguments).wait()) {
                    return true;
                }

                _this.$errors.fail("Unable to execute command '%s'. Use '$ %s %s --help' for help.", beautifiedName, _this.$staticConfig.CLIENT_NAME, beautifiedName);
                return false;
            }

            _this.$logger.fatal("Unknown command '%s'. Use '%s help' for help.", beautifiedName, _this.$staticConfig.CLIENT_NAME);
            _this.tryMatchCommand(commandName);

            return false;
        }).future()();
    };

    CommandsService.prototype.validateMandatoryParams = function (commandArguments, mandatoryParams) {
        var _this = this;
        return (function () {
            var commandArgsHelper = new CommandArgumentsValidationHelper(true, commandArguments);

            if (mandatoryParams.length > 0) {
                if (mandatoryParams.length > commandArguments.length) {
                    _this.$errors.fail("You need to provide all the required parameters.");
                }

                _.each(mandatoryParams, function (mandatoryParam) {
                    var argument = _.first(_.select(commandArgsHelper.remainingArguments, function (c) {
                        return mandatoryParam.validate(c).wait();
                    }));

                    if (argument) {
                        commandArgsHelper.remainingArguments = _.without(commandArgsHelper.remainingArguments, argument);
                    } else {
                        _this.$errors.fail("Missing mandatory parameter.");
                    }
                });
            }

            return commandArgsHelper;
        }).future()();
    };

    CommandsService.prototype.validateCommandArguments = function (command, commandArguments) {
        var _this = this;
        return (function () {
            var mandatoryParams = _.filter(command.allowedParameters, function (param) {
                return param.mandatory;
            });
            var commandArgsHelper = _this.validateMandatoryParams(commandArguments, mandatoryParams).wait();
            if (!commandArgsHelper.isValid) {
                return false;
            }

            if (!command.allowedParameters || command.allowedParameters.length === 0) {
                if (commandArguments.length > 0) {
                    _this.$errors.fail("This command doesn't accept parameters.");
                }
            } else {
                var unverifiedAllowedParams = command.allowedParameters.filter(function (param) {
                    return !param.mandatory;
                });

                _.each(commandArgsHelper.remainingArguments, function (argument) {
                    var parameter = _.find(unverifiedAllowedParams, function (c) {
                        return c.validate(argument).wait();
                    });
                    if (parameter) {
                        unverifiedAllowedParams = _.without(unverifiedAllowedParams, parameter);
                    } else {
                        _this.$errors.fail("The parameter %s is not valid for this command.", argument);
                    }
                });
            }

            return true;
        }).future()();
    };

    CommandsService.prototype.tryMatchCommand = function (commandName) {
        var _this = this;
        var allCommands = this.allCommands(false);
        var similarCommands = [];
        _.each(allCommands, function (command) {
            if (!_this.$injector.isDefaultCommand(command)) {
                command = helpers.stringReplaceAll(command, "|", " ");
                var distance = jaroWinklerDistance(commandName, command);
                if (commandName.length > 3 && command.indexOf(commandName) != -1) {
                    similarCommands.push({ rating: 1, name: command });
                } else if (distance >= 0.65) {
                    similarCommands.push({ rating: distance, name: command });
                }
            }
        });

        similarCommands = _.sortBy(similarCommands, function (command) {
            return -command.rating;
        }).slice(0, 5);

        if (similarCommands.length > 0) {
            var message = ["Did you mean?"];
            _.each(similarCommands, function (command) {
                message.push("\t" + command.name);
            });
            this.$logger.fatal(message.join("\n"));
        }
    };

    CommandsService.prototype.completeCommand = function (commandsWithPlatformArgument, platforms, getPropSchemaAction) {
        var _this = this;
        return (function () {
            var completeCallback = function (err, data) {
                if (err || !data) {
                    return;
                }

                var childrenCommands = _this.$injector.getChildrenCommandsNames(data.prev);

                if (data.words == 1) {
                    return tabtab.log(_this.allCommands(false), data);
                }

                if (data.last.startsWith("--")) {
                    return tabtab.log(Object.keys(require("./options").knownOpts), data, "--");
                }

                if (_.contains(commandsWithPlatformArgument, data.prev)) {
                    return tabtab.log(platforms, data);
                }

                if (data.words == 2 && childrenCommands) {
                    return tabtab.log(_.reject(childrenCommands, function (children) {
                        return children[0] === '*';
                    }), data);
                }

                var propSchema = getPropSchemaAction ? getPropSchemaAction() : null;

                if (propSchema) {
                    var propertyCommands = ["print", "set", "add", "del"];
                    var parseResult = /prop ([^ ]+) ([^ ]*)/.exec(data.line);
                    if (parseResult) {
                        if (_.contains(propertyCommands, parseResult[1])) {
                            var propName = parseResult[2];
                            if (propSchema[propName]) {
                                var range = propSchema[propName].range;
                                if (range) {
                                    if (!_.isArray(range)) {
                                        range = _.map(range, function (value, key) {
                                            return value.input || key;
                                        });
                                    }
                                    return tabtab.log(range, data);
                                }
                            } else {
                                return tabtab.log(Object.keys(propSchema), data);
                            }
                        }
                    }
                }

                return false;
            };

            var tabtab = require("tabtab");
            tabtab.complete(_this.$staticConfig.CLIENT_NAME.toLowerCase(), completeCallback);

            if (_this.$staticConfig.CLIENT_NAME_ALIAS) {
                tabtab.complete(_this.$staticConfig.CLIENT_NAME_ALIAS.toLowerCase(), completeCallback);
            }

            return true;
        }).future()();
    };

    CommandsService.prototype.beautifyCommandName = function (commandName) {
        if (commandName.indexOf("*") > 0) {
            return commandName.substr(0, commandName.indexOf("|"));
        }

        return commandName;
    };
    return CommandsService;
})();
exports.CommandsService = CommandsService;
$injector.register("commandsService", CommandsService);
